import torch
import torch.nn as nn

class DenseGraphAttentionHead(nn.Module):
    """
    A graph attention head.
    Speedup techniques taken from:
    https://github.com/Diego999/pyGAT/blob/master/layers.py
    """
    def __init__(self, input_dim, output_dim):
        super(DenseGraphAttentionHead, self).__init__()
        alpha = 0.2
        self.W = nn.Linear(input_dim, output_dim)
        self.a1 = nn.Linear(output_dim, 1)
        self.a2 = nn.Linear(output_dim, 1)
        self.act = nn.LeakyReLU(alpha)

    def forward(self, nodes, edge_mat):
        Wh = self.W(nodes)
        Wh1 = self.a1(Wh)
        Wh2 = self.a2(Wh)
        attention_mat = self.act(Wh1 + Wh2.T)
        masked_attention_mat = attention_mat.masked_fill_(~edge_mat, float("-inf"))
        attention_final = masked_attention_mat.softmax(dim=1)
        new_node_values = torch.mm(attention_final, Wh)

        return new_node_values

class DenseGraphAttentionLayer(nn.Module):
    """
    A dense layer of the graph attention network.
    """
    def __init__(self, heads, input_dim, intermediate_dim, output_dim):
        super(DenseGraphAttentionLayer, self).__init__()
        self.heads = nn.ModuleList()
        self.final_layer = nn.Linear(heads*intermediate_dim, output_dim)
        for head in range(heads):
            self.heads.extend([DenseGraphAttentionHead(input_dim, intermediate_dim)])

    def forward(self, nodes, edge_mat):
        head_outputs = []
        for head in self.heads:
            head_outputs.append(head(nodes, edge_mat))
        out = torch.cat(head_outputs, dim=-1)
        out = self.final_layer(out).softmax(dim=-1)
        return out

class DenseGraphAttentionNetwork(nn.Module):
    """
    Graph attention for Dense Graphs, where the edge list
    is given by an NxN array, where it is True if there is an
    edge from (i,j), False otherwise.

    Assumes all nodes are self connected (the diagonal has all values as `true`)
    """
    def __init__(self, depth, heads, input_dim, inner_dim):
        super(DenseGraphAttentionNetwork, self).__init__()
        self.layers = nn.ModuleList()
        if depth == 1:
            inner_dim = input_dim

        for layer_no in range(depth):
            if layer_no == 0:
                self.layers.extend([DenseGraphAttentionLayer(heads, input_dim, inner_dim, inner_dim)])
            elif layer_no == depth-1:
                self.layers.extend([DenseGraphAttentionLayer(heads, inner_dim, inner_dim, input_dim)])
            else:
                self.layers.extend([DenseGraphAttentionLayer(heads, inner_dim, inner_dim, inner_dim)])

    def forward(self, nodes, edge_mat):
        for layer in self.layers:
            nodes = layer(nodes, edge_mat)
        return nodes



import torch
import torch.nn as nn

class GraphAttentionHead(nn.Module):
    """
    A single head for the graph attention network.
    Currently, this implementation is slow because of the
    for loop.
    """
    def __init__(self, input_dim, output_dim, slope=0.2):
        super(GraphAttentionHead, self).__init__()
        self.w = nn.Linear(input_dim, output_dim)
        self.a = nn.Linear(2*output_dim, 1)
        self.activation = nn.LeakyReLU(slope)

    def forward(self, nodes, edges):
        nodes = self.w(nodes)
        new_nodes = nodes
        for curr_node, neighbors in edges.items():
            neighbors.extend([curr_node]) # Don't forget self attention!
            top_list = torch.tensor([self._get_top(nodes, curr_node, neighbor) for neighbor in neighbors])
            new_nodes[curr_node] = self._get_new_node_info(nodes, neighbors, top_list)
        return new_nodes

    def _get_top(self, nodes, i, j):
        """
        Returns a scalar for the top of the a_ij function
        nodes: Info of the nodes after the W param matrix
        i,j: Nodes to consider
        """
        cat = self.a(torch.cat((nodes[i], nodes[j]), dim=-1))
        x = torch.exp(self.activation(cat))
        return x

    def _get_new_node_info(self, nodes, neighbors, top_list):
        """
        nodes: List of nodes after the conv
        neighbors: list of which nodes are the neighbors of the current one
        top_list: list of a_ij from the function
        """
        bottom = torch.sum(top_list)
        x = torch.stack([top_list[curr_neighbor_no] * nodes[curr_neighbor] for curr_neighbor_no, curr_neighbor in enumerate(neighbors)])/bottom
        return torch.sum(x)

class GraphAttentionLayer(nn.Module):
    """
    A layer of the Graph Attention Network.
    NOTE! This is a different implementation than the original
    paper. Instead of averaging, this applies a learnable layer of
    size KFxF, more closely resembling the original attention paper.
    """
    def __init__(self, heads, input_dim, intermediate_dim, output_dim):
        super(GraphAttentionLayer, self).__init__()
        self.heads = nn.ModuleList()
        self.final_layer = nn.Linear(heads*intermediate_dim, output_dim)
        for head_no in range(heads):
            self.heads.extend([GraphAttentionHead(input_dim, intermediate_dim)])

    def forward(self, nodes, edges):
        head_outputs = []
        for head in self.heads:
            head_outputs.append(head(nodes, edges))
        out = torch.cat(head_outputs, dim=-1)
        out = self.final_layer(out).softmax(dim=-1)
        return out

class GraphAttentionNetwork(nn.Module):
    """
    The entrance for the Graph attention Network.
    This assumes that the inputs are:
    nodes: An NxF array, with N nodes and F dimensions for the features
    edges: A dictonary that represents directed edges. In particular, it has the form
    {x: [y,z]}
    which means that x has an edge to y and z. The edges do not repeat.
    Also, the edges are expected to be 0 indexed and in integer form.

    This function just returns the nodes, as the edges are unchanged.
    """
    def __init__(self, depth, heads, input_dim, inner_dim):
        super(GraphAttentionNetwork, self).__init__()
        self.layers = nn.ModuleList()
        if depth == 1:
            inner_dim = input_dim

        for layer_no in range(depth):
            if layer_no == 0:
                self.layers.extend([GraphAttentionLayer(heads, input_dim, inner_dim, inner_dim)])
            elif layer_no == depth-1:
                self.layers.extend([GraphAttentionLayer(heads, inner_dim, inner_dim, input_dim)])
            else:
                self.layers.extend([GraphAttentionLayer(heads, inner_dim, inner_dim, inner_dim)])

    def forward(self, nodes, edges):
        for layer in self.layers:
            nodes = layer(nodes, edges)
        return nodes



import torch
import torch.nn as nn

class GraphUtils:
    """
    This class provides an easy way to convert a
    graph's edges from an adjacency list data store to an
    adjacency matrix and back again.
    """
    @staticmethod
    def list_to_matrix(num_nodes, edge_list):
        mat = torch.eye(num_nodes, dtype=torch.bool)
        for start, dests in edge_list.items():
            for dest in dests:
                mat[start, dest] = True
        return mat

    @staticmethod
    def matrix_to_list(edge_matrix):
        l = dict()
        for i, neighbors in enumerate(edge_matrix):
            l[i] = set(map(lambda tensor: tensor.item(), torch.nonzero(neighbors).flatten()))
            l[i].discard(i)
        return l



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fast_gat",
    version="0.1.1",
    author="Peter Tatkowski",
    author_email="tatp22@gmail.com",
    description="A PyTorch implementation of Graph Attention Networks, with experimental speedup features.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tatp22/pytorch-fast-GAT",
    packages=setuptools.find_packages(),
    keywords=['attention', 'deep learning', 'artificial intelligence', 'sparse attention'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    install_requires=[
        'torch'
    ],
)
